from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any

from .jsonschema_rs import (
    Draft4,
    Draft4Validator,
    Draft6,
    Draft6Validator,
    Draft7,
    Draft7Validator,
    Draft201909,
    Draft201909Validator,
    Draft202012,
    Draft202012Validator,
    EmailOptions,
    Evaluation,
    FancyRegexOptions,
    HttpOptions,
    RegexOptions,
    Registry,
    ValidationErrorKind,
    canonical,
    evaluate,
    is_valid,
    iter_errors,
    meta,
    validate,
    validator_cls_for,
    bundle,
    validator_for,
)

Validator = Draft4Validator | Draft6Validator | Draft7Validator | Draft201909Validator | Draft202012Validator


class ValidationError(ValueError):
    """An instance is invalid under a provided schema."""

    message: str
    verbose_message: str
    schema_path: list[str | int]
    instance_path: list[str | int]
    evaluation_path: list[str | int]
    kind: ValidationErrorKind
    instance: Any
    absolute_keyword_location: str | None

    def __init__(
        self,
        message: str,
        verbose_message: str,
        schema_path: list[str | int],
        instance_path: list[str | int],
        evaluation_path: list[str | int],
        kind: ValidationErrorKind,
        instance: Any,
        absolute_keyword_location: str | None = None,
    ) -> None:
        super().__init__(verbose_message)
        self.message = message
        self.verbose_message = verbose_message
        self.schema_path = schema_path
        self.instance_path = instance_path
        self.evaluation_path = evaluation_path
        self.kind = kind
        self.instance = instance
        self.absolute_keyword_location = absolute_keyword_location

    def __str__(self) -> str:
        return self.verbose_message

    def __repr__(self) -> str:
        return f"<ValidationError: '{self.message}'>"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ValidationError):
            return NotImplemented
        return (
            self.message == other.message
            and self.schema_path == other.schema_path
            and self.instance_path == other.instance_path
        )

    def __hash__(self) -> int:
        return hash((self.message, tuple(self.schema_path), tuple(self.instance_path)))


class ReferencingError(Exception):
    """Errors that can occur during reference resolution and resource handling."""

    message: str

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message

    def __str__(self) -> str:
        return self.message

    def __repr__(self) -> str:
        return f"<ReferencingError: '{self.message}'>"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ReferencingError):
            return NotImplemented
        return self.message == other.message

    def __hash__(self) -> int:
        return hash(self.message)


__all__ = [
    "ReferencingError",
    "ValidationError",
    "canonical",
    "ValidationErrorKind",
    "Evaluation",
    "bundle",
    "is_valid",
    "validate",
    "iter_errors",
    "evaluate",
    "validator_cls_for",
    "validator_for",
    "Draft4",
    "Draft6",
    "Draft7",
    "Draft201909",
    "Draft202012",
    "Draft4Validator",
    "Draft6Validator",
    "Draft7Validator",
    "Draft201909Validator",
    "Draft202012Validator",
    "Validator",
    "Registry",
    "EmailOptions",
    "FancyRegexOptions",
    "HttpOptions",
    "RegexOptions",
    "meta",
]

del TYPE_CHECKING, annotations  # noqa: F821



import pytest

import jsonschema_rs
from jsonschema_rs import (
    Draft4Validator,
    Draft6Validator,
    Draft7Validator,
    Draft201909Validator,
    Draft202012Validator,
    validator_cls_for,
)


@pytest.mark.parametrize(
    ["schema", "expected_cls"],
    [
        # No $schema defaults to Draft202012
        ({"type": "string"}, Draft202012Validator),
        ({}, Draft202012Validator),
        # Boolean schemas (no $schema) → Draft202012
        (True, Draft202012Validator),
        (False, Draft202012Validator),
        # Draft 4 - both HTTP and HTTPS, with and without trailing #
        ({"$schema": "http://json-schema.org/draft-04/schema#"}, Draft4Validator),
        ({"$schema": "https://json-schema.org/draft-04/schema#"}, Draft4Validator),
        ({"$schema": "http://json-schema.org/draft-04/schema"}, Draft4Validator),
        # Draft 6
        ({"$schema": "http://json-schema.org/draft-06/schema#"}, Draft6Validator),
        ({"$schema": "https://json-schema.org/draft-06/schema#"}, Draft6Validator),
        # Draft 7
        ({"$schema": "http://json-schema.org/draft-07/schema#"}, Draft7Validator),
        ({"$schema": "https://json-schema.org/draft-07/schema#"}, Draft7Validator),
        # Draft 2019-09
        ({"$schema": "https://json-schema.org/draft/2019-09/schema"}, Draft201909Validator),
        ({"$schema": "http://json-schema.org/draft/2019-09/schema"}, Draft201909Validator),
        # Draft 2020-12
        ({"$schema": "https://json-schema.org/draft/2020-12/schema"}, Draft202012Validator),
        ({"$schema": "http://json-schema.org/draft/2020-12/schema"}, Draft202012Validator),
        # Unknown $schema falls back to Draft202012
        ({"$schema": "http://custom.example.com/schema"}, Draft202012Validator),
    ],
)
def test_draft_detection(schema, expected_cls):
    assert validator_cls_for(schema) is expected_cls


def test_returned_class_creates_working_validator():
    cls = validator_cls_for({"type": "string"})
    validator = cls({"type": "string"})
    assert validator.is_valid("hello")
    assert not validator.is_valid(42)


def test_no_meta_validation():
    # validator_cls_for performs no meta-validation — only draft detection.
    # Invalid schemas do not raise here; they raise only when cls(schema) is called.
    schema = {"$schema": "http://json-schema.org/draft-07/schema#", "type": "invalid_type"}
    cls = validator_cls_for(schema)
    assert cls is Draft7Validator
    with pytest.raises(jsonschema_rs.ValidationError):
        cls(schema)



from .. import JSONType

def to_string(object: JSONType) -> str:
    """Serialize a Python object to a canonical JSON string.

    Main use case: deduplicating equivalent JSON Schemas.

    - Dict keys are sorted lexicographically (byte order).
    - Integer-valued floats are serialized as integers (1.0 → 1).
    - NaN and Infinity are serialized as null.
    - Output is always compact.

    Raises ValueError on serialization failure (e.g., recursion limit exceeded,
    unsupported type, lone surrogates in strings).
    """
    ...



import jsonschema_rs
import pytest

PERSON_SCHEMA = {
    "$id": "https://example.com/person.json",
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "type": "object",
    "properties": {"name": {"type": "string"}},
    "required": ["name"],
}


def test_bundle_no_external_refs():
    bundled = jsonschema_rs.bundle({"type": "string"})
    assert isinstance(bundled, dict)
    assert "$defs" not in bundled


def test_bundle_single_external_ref():
    root = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$ref": "https://example.com/person.json",
    }
    registry = jsonschema_rs.Registry(
        resources=[("https://example.com/person.json", PERSON_SCHEMA)]
    )
    bundled = jsonschema_rs.bundle(root, registry=registry)
    assert bundled["$ref"] == "https://example.com/person.json"  # must not be rewritten
    assert "https://example.com/person.json" in bundled["$defs"]


def test_bundle_validates_identically():
    root = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$ref": "https://example.com/person.json",
    }
    registry = jsonschema_rs.Registry(
        resources=[("https://example.com/person.json", PERSON_SCHEMA)]
    )
    bundled = jsonschema_rs.bundle(root, registry=registry)
    validator = jsonschema_rs.validator_for(bundled)
    assert validator.is_valid({"name": "Alice"})
    assert not validator.is_valid({"age": 30})


def test_bundle_unresolvable_raises():
    with pytest.raises(jsonschema_rs.ReferencingError):
        jsonschema_rs.bundle({"$ref": "https://example.com/missing.json"})


def test_bundle_resolves_ref_with_nested_id_scope():
    root = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$defs": {"A": {"$id": "https://example.com/A/", "$ref": "b.json"}},
    }
    registry = jsonschema_rs.Registry(
        resources=[("https://example.com/A/b.json", {"type": "integer"})]
    )
    bundled = jsonschema_rs.bundle(root, registry=registry)
    assert "A" in bundled["$defs"]
    assert "https://example.com/A/b.json" in bundled["$defs"]


def test_bundle_ignores_ref_inside_const_annotation_payload():
    schema = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "const": {"$ref": "https://example.com/not-a-schema"},
    }
    bundled = jsonschema_rs.bundle(schema)
    assert bundled == schema
    assert "$defs" not in bundled


def test_bundle_supports_legacy_drafts():
    resource_uri = "https://example.com/legacy/person.json"
    for schema_uri in (
        "http://json-schema.org/draft-04/schema#",
        "http://json-schema.org/draft-06/schema#",
        "http://json-schema.org/draft-07/schema#",
    ):
        root = {"$schema": schema_uri, "$ref": resource_uri}
        registry = jsonschema_rs.Registry(
            resources=[
                (
                    resource_uri,
                    {"$schema": schema_uri, "type": "integer", "minimum": 0},
                )
            ]
        )
        distributed = jsonschema_rs.validator_for(root, registry=registry)
        bundled = jsonschema_rs.bundle(root, registry=registry)
        bundled_validator = jsonschema_rs.validator_for(bundled)

        assert "$defs" not in bundled
        assert resource_uri in bundled["definitions"]
        embedded = bundled["definitions"][resource_uri]
        assert embedded.get("id") == resource_uri or embedded.get("$id") == resource_uri

        for instance in (0, 5):
            assert distributed.is_valid(instance)
            assert bundled_validator.is_valid(instance)
        for instance in (-1, "x", 1.5):
            assert not distributed.is_valid(instance)
            assert not bundled_validator.is_valid(instance)


def test_bundle_mixed_draft_refs_validate_identically():
    resource_uri = "https://example.com/mixed/schema.json"
    root = {"$schema": "http://json-schema.org/draft-07/schema#", "$ref": resource_uri}
    registry = jsonschema_rs.Registry(
        resources=[
            (
                resource_uri,
                {"$schema": "http://json-schema.org/draft-04/schema#", "type": "integer"},
            )
        ]
    )

    distributed = jsonschema_rs.validator_for(root, registry=registry)
    bundled = jsonschema_rs.bundle(root, registry=registry)
    bundled_validator = jsonschema_rs.validator_for(bundled)

    embedded = bundled["definitions"][resource_uri]
    assert embedded["id"] == resource_uri
    assert embedded["$id"] == resource_uri
    assert distributed.is_valid(1)
    assert bundled_validator.is_valid(1)
    assert not distributed.is_valid("x")
    assert not bundled_validator.is_valid("x")


def test_bundle_mixed_draft_const_semantics_are_preserved():
    resource_uri = "https://example.com/mixed/const.json"
    root = {"$schema": "http://json-schema.org/draft-04/schema#", "$ref": resource_uri}
    registry = jsonschema_rs.Registry(
        resources=[
            (
                resource_uri,
                {"$schema": "http://json-schema.org/draft-07/schema#", "const": 1},
            )
        ]
    )

    distributed = jsonschema_rs.validator_for(root, registry=registry)
    bundled = jsonschema_rs.bundle(root, registry=registry)
    bundled_validator = jsonschema_rs.validator_for(bundled)

    assert distributed.is_valid(1)
    assert bundled_validator.is_valid(1)
    assert not distributed.is_valid(2)
    assert not bundled_validator.is_valid(2)



import enum
import json
import sys
from decimal import Decimal

import pytest
from hypothesis import given
from hypothesis import strategies as st

from jsonschema_rs import canonical

to_string = canonical.json.to_string

# Recursive strategy for JSON-compatible values (no NaN/Inf — those don't roundtrip)
_json_scalars = st.one_of(
    st.none(),
    st.booleans(),
    st.integers(),
    st.floats(allow_nan=False, allow_infinity=False),
    st.text(),
)
_json_values = st.recursive(
    _json_scalars,
    lambda children: st.one_of(
        st.lists(children),
        st.dictionaries(st.text(), children),
    ),
)


class Color(enum.Enum):
    RED = "red"
    GREEN = 1


class StrColor(str, enum.Enum):
    BLUE = "blue"


@pytest.mark.parametrize(
    "value, expected",
    [
        (None, "null"),
        (True, "true"),
        (False, "false"),
        (0, "0"),
        (42, "42"),
        (-7, "-7"),
        ("hello", '"hello"'),
        ("", '""'),
        (1.0, "1"),
        (0.0, "0"),
        (-5.0, "-5"),
        (1.5, "1.5"),
        (float("nan"), "null"),
        (float("inf"), "null"),
        (float("-inf"), "null"),
        (2**128, str(2**128)),
        (-(2**128), str(-(2**128))),
        ({"b": 1, "a": 2}, '{"a":2,"b":1}'),
        ({"a": 1, "b": 2}, '{"a":1,"b":2}'),
        ({}, "{}"),
        ({"x": 99}, '{"x":99}'),
        ({"z": {"b": 1, "a": 2}, "a": 0}, '{"a":0,"z":{"a":2,"b":1}}'),
        ([], "[]"),
        ([1, 2, 3], "[1,2,3]"),
        ([{"b": 1, "a": 2}], '[{"a":2,"b":1}]'),
        ((), "[]"),
        ((1, 2, 3), "[1,2,3]"),
        (Color.RED, '"red"'),
        (Color.GREEN, "1"),
        ({StrColor.BLUE: "sky"}, '{"blue":"sky"}'),
        (Decimal("1.0"), "1"),
        (Decimal("100"), "100"),
        (Decimal("NaN"), "null"),
        (Decimal("Infinity"), "null"),
        (Decimal("-Infinity"), "null"),
        (Decimal(2**128), str(2**128)),
    ],
)
def test_to_string(value, expected):
    assert to_string(value) == expected


def test_float_large_integer_valued():
    result = to_string(1e300)
    assert result == str(int(1e300))


@pytest.mark.parametrize("value", [float(2**63), float(-(2**63)), float(2**64)])
def test_float_integer_boundary_values(value):
    assert to_string(value) == str(int(value))


def test_decimal_fractional():
    assert json.loads(to_string(Decimal("1.5"))) == pytest.approx(1.5)


@pytest.mark.parametrize("value, expected", [
    (Decimal("100E-2"), "1"),
    (Decimal("1E-2"), "0.01"),
    (Decimal("0E-1000"), "0"),
    (Decimal("-0E-1000"), "0"),
])
def test_decimal_exponent_integrality(value, expected):
    assert to_string(value) == expected


@pytest.mark.parametrize("value", [object(), {1, 2, 3}])
def test_unsupported_type_raises(value):
    with pytest.raises(ValueError):
        to_string(value)


@pytest.mark.parametrize(
    "value",
    [
        None,
        True,
        False,
        0,
        42,
        -7,
        "hello",
        1.5,
        [1, 2, 3],
        {"b": 1, "a": 2},
        {"nested": {"z": 0, "a": 1}},
    ],
)
def test_roundtrip(value):
    assert json.loads(to_string(value)) == value


def test_same_dict_different_order_produces_same_output():
    assert to_string({"a": 1, "b": 2, "c": 3}) == to_string({"c": 3, "a": 1, "b": 2})


@pytest.mark.parametrize("float_val, int_val", [(1.0, 1), (0.0, 0)])
def test_integer_float_same_as_int(float_val, int_val):
    assert to_string(float_val) == to_string(int_val)


@pytest.mark.parametrize("value", ["\ud800", {"\ud800": 1}])
def test_lone_surrogate_raises(value):
    with pytest.raises(ValueError, match="surrogates not allowed"):
        to_string(value)


def test_str_enum_key_value_lookup_error():
    class BrokenStrEnum(str, enum.Enum):
        A = "a"

        def __getattribute__(self, name):
            if name == "value":
                raise RuntimeError("boom")
            return super().__getattribute__(name)

    with pytest.raises(ValueError, match="Failed to access enum key value"):
        to_string({BrokenStrEnum.A: 1})


def test_enum_value_lookup_error():
    class BrokenEnum(enum.Enum):
        A = "a"

        def __getattribute__(self, name):
            if name == "value":
                raise RuntimeError("boom")
            return super().__getattribute__(name)

    with pytest.raises(ValueError, match="boom"):
        to_string(BrokenEnum.A)


@given(_json_values)
def test_roundtrip_hypothesis(value):
    # json.loads(to_string(v)) == v holds for all JSON-compatible values.
    # Python's == handles int/float equivalence (1 == 1.0), so integer-valued
    # floats that become ints after the round-trip still compare equal.
    assert json.loads(to_string(value)) == value


@given(_json_values)
def test_idempotent(value):
    # Re-encoding the parsed output produces the same string.
    first = to_string(value)
    assert to_string(json.loads(first)) == first


@given(st.dictionaries(st.text(), _json_values))
def test_dict_key_order_invariance(d):
    reversed_d = dict(reversed(list(d.items())))
    assert to_string(d) == to_string(reversed_d)


@given(st.integers())
def test_integer_float_equivalence(n):
    f = float(n)
    if f == n:  # skip integers outside the exact float range
        assert to_string(f) == to_string(n)


@pytest.mark.skipif(not hasattr(sys, "getrefcount"), reason="PyPy does not have sys.getrefcount")
def test_enum_value_refcount_is_stable():
    class PayloadEnum(enum.Enum):
        ITEM = [1]

    payload = PayloadEnum.ITEM.value
    baseline = sys.getrefcount(payload)
    for _ in range(200):
        assert to_string(PayloadEnum.ITEM) == "[1]"
    assert sys.getrefcount(payload) == baseline



from collections.abc import Iterator
from decimal import Decimal
from typing import Any, Callable, List, Protocol, TypeAlias, TypeVar, TypedDict, Union

from . import canonical as canonical

_SchemaT = TypeVar("_SchemaT", bool, dict[str, Any])
_FormatFunc = TypeVar("_FormatFunc", bound=Callable[[str], bool])
JSONType: TypeAlias = dict[str, Any] | list | str | int | float | Decimal | bool | None
JSONPrimitive: TypeAlias = str | int | float | Decimal | bool | None

class KeywordValidator(Protocol):
    """Protocol for custom keyword validators.

    Custom keywords are classes instantiated with (parent_schema, value, schema_path)
    that implement a validate(instance) method which raises an exception on failure.

    Example:
        class DivisibleBy:
            def __init__(self, parent_schema, value, schema_path):
                self.divisor = value

            def validate(self, instance):
                if isinstance(instance, int) and instance % self.divisor != 0:
                    raise ValueError(f"{instance} is not divisible by {self.divisor}")

        validator = jsonschema_rs.validator_for(
            {"divisibleBy": 3},
            keywords={"divisibleBy": DivisibleBy},
        )

    """

    def __init__(self, parent_schema: dict[str, Any], value: Any, schema_path: list[str | int]) -> None: ...
    def validate(self, instance: JSONType) -> None: ...

class EvaluationAnnotation(TypedDict):
    schemaLocation: str
    absoluteKeywordLocation: str | None
    instanceLocation: str
    annotations: JSONType

class EvaluationErrorEntry(TypedDict):
    schemaLocation: str
    absoluteKeywordLocation: str | None
    instanceLocation: str
    error: str

class FlagOutput(TypedDict):
    """JSON Schema Output v1 - Flag format."""

    valid: bool

class OutputUnit(TypedDict, total=False):
    """A single output unit in list/hierarchical formats."""

    valid: bool
    evaluationPath: str
    schemaLocation: str
    instanceLocation: str
    errors: dict[str, str]
    annotations: JSONType
    droppedAnnotations: JSONType
    details: List["OutputUnit"]

class ListOutput(TypedDict):
    """JSON Schema Output v1 - List format."""

    valid: bool
    details: List[OutputUnit]

class Evaluation:
    valid: bool
    def flag(self) -> FlagOutput: ...
    def list(self) -> ListOutput: ...
    def hierarchical(self) -> OutputUnit: ...
    def annotations(self) -> List[EvaluationAnnotation]: ...
    def errors(self) -> List[EvaluationErrorEntry]: ...
    def __repr__(self) -> str: ...

class FancyRegexOptions:
    def __init__(
        self, backtrack_limit: int | None = None, size_limit: int | None = None, dfa_size_limit: int | None = None
    ) -> None: ...
    def __repr__(self) -> str: ...

class RegexOptions:
    def __init__(self, size_limit: int | None = None, dfa_size_limit: int | None = None) -> None: ...
    def __repr__(self) -> str: ...

class EmailOptions:
    """Configuration for email format validation."""

    def __init__(
        self,
        require_tld: bool = False,
        allow_domain_literal: bool = True,
        allow_display_text: bool = True,
        minimum_sub_domains: int | None = None,
    ) -> None: ...
    def __repr__(self) -> str: ...

class HttpOptions:
    """Configuration for HTTP client used in schema retrieval."""

    timeout: float | None
    connect_timeout: float | None
    tls_verify: bool
    ca_cert: str | None

    def __init__(
        self,
        timeout: float | None = None,
        connect_timeout: float | None = None,
        tls_verify: bool = True,
        ca_cert: str | None = None,
    ) -> None: ...
    def __repr__(self) -> str: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...

PatternOptionsType = Union[FancyRegexOptions, RegexOptions]

class RetrieverProtocol(Protocol):
    def __call__(self, uri: str) -> JSONType: ...

def is_valid(
    schema: _SchemaT,
    instance: Any,
    draft: int | None = None,
    formats: dict[str, _FormatFunc] | None = None,
    validate_formats: bool | None = None,
    ignore_unknown_formats: bool = True,
    retriever: RetrieverProtocol | None = None,
    registry: Registry | None = None,
    mask: str | None = None,
    base_uri: str | None = None,
    pattern_options: PatternOptionsType | None = None,
    email_options: EmailOptions | None = None,
    http_options: HttpOptions | None = None,
    keywords: dict[str, type[KeywordValidator]] | None = None,
) -> bool:
    """Check if a JSON instance is valid against a schema.

    Returns True if valid, False otherwise. Stops at the first error.
    """
    ...

def validate(
    schema: _SchemaT,
    instance: Any,
    draft: int | None = None,
    formats: dict[str, _FormatFunc] | None = None,
    validate_formats: bool | None = None,
    ignore_unknown_formats: bool = True,
    retriever: RetrieverProtocol | None = None,
    registry: Registry | None = None,
    mask: str | None = None,
    base_uri: str | None = None,
    pattern_options: PatternOptionsType | None = None,
    email_options: EmailOptions | None = None,
    http_options: HttpOptions | None = None,
    keywords: dict[str, type[KeywordValidator]] | None = None,
) -> None:
    """Validate a JSON instance against a schema.

    Raises ValidationError if invalid. Stops at the first error.
    """
    ...

def iter_errors(
    schema: _SchemaT,
    instance: Any,
    draft: int | None = None,
    formats: dict[str, _FormatFunc] | None = None,
    validate_formats: bool | None = None,
    ignore_unknown_formats: bool = True,
    retriever: RetrieverProtocol | None = None,
    registry: Registry | None = None,
    mask: str | None = None,
    base_uri: str | None = None,
    pattern_options: PatternOptionsType | None = None,
    email_options: EmailOptions | None = None,
    http_options: HttpOptions | None = None,
    keywords: dict[str, type[KeywordValidator]] | None = None,
) -> Iterator[ValidationError]:
    """Iterate over all validation errors.

    Returns an iterator of ValidationError objects for all errors found.
    """
    ...

def evaluate(
    schema: _SchemaT,
    instance: Any,
    draft: int | None = None,
    formats: dict[str, _FormatFunc] | None = None,
    validate_formats: bool | None = None,
    ignore_unknown_formats: bool = True,
    retriever: RetrieverProtocol | None = None,
    registry: Registry | None = None,
    base_uri: str | None = None,
    pattern_options: PatternOptionsType | None = None,
    email_options: EmailOptions | None = None,
    http_options: HttpOptions | None = None,
    keywords: dict[str, type[KeywordValidator]] | None = None,
) -> Evaluation:
    """Evaluate an instance and return structured output.

    Returns an Evaluation object with flag(), list(), and hierarchical() output formats.
    """
    ...

class ReferencingError:
    message: str

class ValidationErrorKind:
    @property
    def name(self) -> str:
        """The JSON Schema keyword that triggered this error."""
        ...
    @property
    def value(self) -> Any: ...
    def as_dict(self) -> dict[str, Any]: ...

    class AdditionalItems:
        limit: int

    class AdditionalProperties:
        unexpected: list[str]

    class AnyOf:
        context: list[list["ValidationError"]]

    class BacktrackLimitExceeded:
        error: str

    class Constant:
        expected_value: JSONType

    class Contains: ...

    class ContentEncoding:
        content_encoding: str

    class ContentMediaType:
        content_media_type: str

    class Custom:
        keyword: str
        message: str

    class Enum:
        options: list[JSONType]

    class ExclusiveMaximum:
        limit: JSONPrimitive

    class ExclusiveMinimum:
        limit: JSONPrimitive

    class FalseSchema: ...

    class Format:
        format: str

    class FromUtf8:
        error: str

    class MaxItems:
        limit: int

    class Maximum:
        limit: JSONPrimitive

    class MaxLength:
        limit: int

    class MaxProperties:
        limit: int

    class MinItems:
        limit: int

    class Minimum:
        limit: JSONPrimitive

    class MinLength:
        limit: int

    class MinProperties:
        limit: int

    class MultipleOf:
        multiple_of: int | float | Decimal

    class Not:
        schema: JSONType

    class OneOfMultipleValid:
        context: list[list["ValidationError"]]

    class OneOfNotValid:
        context: list[list["ValidationError"]]

    class Pattern:
        pattern: str

    class PropertyNames:
        error: "ValidationError"

    class Required:
        property: str

    class Type:
        types: list[str]

    class UnevaluatedItems:
        unexpected: list[int]

    class UnevaluatedProperties:
        unexpected: list[str]

    class UniqueItems: ...

    class Referencing:
        error: ReferencingError

class ValidationError(ValueError):
    message: str
    verbose_message: str
    schema_path: list[str | int]
    instance_path: list[str | int]
    evaluation_path: list[str | int]
    kind: ValidationErrorKind
    instance: JSONType

Draft4: int
Draft6: int
Draft7: int
Draft201909: int
Draft202012: int

class Draft4Validator:
    def __init__(
        self,
        schema: _SchemaT | str,
        formats: dict[str, _FormatFunc] | None = None,
        validate_formats: bool | None = None,
        ignore_unknown_formats: bool = True,
        retriever: RetrieverProtocol | None = None,
        registry: Registry | None = None,
        mask: str | None = None,
        base_uri: str | None = None,
        pattern_options: PatternOptionsType | None = None,
        email_options: EmailOptions | None = None,
        http_options: HttpOptions | None = None,
        keywords: dict[str, type[KeywordValidator]] | None = None,
    ) -> None: ...
    def is_valid(self, instance: Any) -> bool: ...
    def validate(self, instance: Any) -> None: ...
    def iter_errors(self, instance: Any) -> Iterator[ValidationError]: ...
    def evaluate(self, instance: Any) -> Evaluation: ...
    def __repr__(self) -> str: ...

class Draft6Validator:
    def __init__(
        self,
        schema: _SchemaT | str,
        formats: dict[str, _FormatFunc] | None = None,
        validate_formats: bool | None = None,
        ignore_unknown_formats: bool = True,
        retriever: RetrieverProtocol | None = None,
        registry: Registry | None = None,
        mask: str | None = None,
        base_uri: str | None = None,
        pattern_options: PatternOptionsType | None = None,
        email_options: EmailOptions | None = None,
        http_options: HttpOptions | None = None,
        keywords: dict[str, type[KeywordValidator]] | None = None,
    ) -> None: ...
    def is_valid(self, instance: Any) -> bool: ...
    def validate(self, instance: Any) -> None: ...
    def iter_errors(self, instance: Any) -> Iterator[ValidationError]: ...
    def evaluate(self, instance: Any) -> Evaluation: ...
    def __repr__(self) -> str: ...

class Draft7Validator:
    def __init__(
        self,
        schema: _SchemaT | str,
        formats: dict[str, _FormatFunc] | None = None,
        validate_formats: bool | None = None,
        ignore_unknown_formats: bool = True,
        retriever: RetrieverProtocol | None = None,
        registry: Registry | None = None,
        mask: str | None = None,
        base_uri: str | None = None,
        pattern_options: PatternOptionsType | None = None,
        email_options: EmailOptions | None = None,
        http_options: HttpOptions | None = None,
        keywords: dict[str, type[KeywordValidator]] | None = None,
    ) -> None: ...
    def is_valid(self, instance: Any) -> bool: ...
    def validate(self, instance: Any) -> None: ...
    def iter_errors(self, instance: Any) -> Iterator[ValidationError]: ...
    def evaluate(self, instance: Any) -> Evaluation: ...
    def __repr__(self) -> str: ...

class Draft201909Validator:
    def __init__(
        self,
        schema: _SchemaT | str,
        formats: dict[str, _FormatFunc] | None = None,
        validate_formats: bool | None = None,
        ignore_unknown_formats: bool = True,
        retriever: RetrieverProtocol | None = None,
        registry: Registry | None = None,
        mask: str | None = None,
        base_uri: str | None = None,
        pattern_options: PatternOptionsType | None = None,
        email_options: EmailOptions | None = None,
        http_options: HttpOptions | None = None,
        keywords: dict[str, type[KeywordValidator]] | None = None,
    ) -> None: ...
    def is_valid(self, instance: Any) -> bool: ...
    def validate(self, instance: Any) -> None: ...
    def iter_errors(self, instance: Any) -> Iterator[ValidationError]: ...
    def evaluate(self, instance: Any) -> Evaluation: ...
    def __repr__(self) -> str: ...

class Draft202012Validator:
    def __init__(
        self,
        schema: _SchemaT | str,
        formats: dict[str, _FormatFunc] | None = None,
        validate_formats: bool | None = None,
        ignore_unknown_formats: bool = True,
        retriever: RetrieverProtocol | None = None,
        registry: Registry | None = None,
        mask: str | None = None,
        base_uri: str | None = None,
        pattern_options: PatternOptionsType | None = None,
        email_options: EmailOptions | None = None,
        http_options: HttpOptions | None = None,
        keywords: dict[str, type[KeywordValidator]] | None = None,
    ) -> None: ...
    def is_valid(self, instance: Any) -> bool: ...
    def validate(self, instance: Any) -> None: ...
    def iter_errors(self, instance: Any) -> Iterator[ValidationError]: ...
    def evaluate(self, instance: Any) -> Evaluation: ...
    def __repr__(self) -> str: ...

Validator: TypeAlias = Draft4Validator | Draft6Validator | Draft7Validator | Draft201909Validator | Draft202012Validator

def validator_for(
    schema: _SchemaT,
    formats: dict[str, _FormatFunc] | None = None,
    validate_formats: bool | None = None,
    ignore_unknown_formats: bool = True,
    retriever: RetrieverProtocol | None = None,
    registry: Registry | None = None,
    mask: str | None = None,
    base_uri: str | None = None,
    pattern_options: PatternOptionsType | None = None,
    email_options: EmailOptions | None = None,
    http_options: HttpOptions | None = None,
    keywords: dict[str, type[KeywordValidator]] | None = None,
) -> Validator:
    """Create a validator for the given schema.

    Automatically detects the JSON Schema draft from the $schema keyword.
    Returns a Draft-specific validator instance.
    """
    ...

def bundle(
    schema: Any,
    /,
    *,
    retriever: RetrieverProtocol | None = None,
    registry: Registry | None = None,
    draft: int | None = None,
    base_uri: str | None = None,
) -> dict[str, Any]:
    """Bundle a JSON Schema into a Compound Schema Document.

    All externally-referenced schemas reachable via $ref are embedded in a
    draft-appropriate container (definitions for Draft 4/6/7, $defs for
    Draft 2019-09/2020-12).
    Original $ref values are preserved unchanged.
    """
    ...

def validator_cls_for(schema: _SchemaT) -> type[Validator]:
    """Detect the JSON Schema draft for a schema and return the corresponding validator class.

    Draft is detected automatically from the $schema field. Defaults to Draft202012Validator.
    """
    ...

class Registry:
    def __init__(
        self,
        resources: list[tuple[str, JSONType]],
        draft: int | None = None,
        retriever: RetrieverProtocol | None = None,
    ) -> None: ...
    def __repr__(self) -> str: ...

class _Meta:
    def is_valid(self, schema: _SchemaT, registry: Registry | None = None) -> bool: ...
    def validate(self, schema: _SchemaT, registry: Registry | None = None) -> None: ...

meta: _Meta



from .. import JSONType


def clone(object: JSONType) -> JSONType:
    """Deep-clone a JSON-compatible Python object.

    Only ``dict`` and ``list`` are copied. All other values (str, int, float,
    bool, None) are shared by reference since they are immutable.

    Dict subclasses (e.g. ``CaseInsensitiveDict``) are cloned into plain
    ``dict``. All other non-JSON types are returned as-is (treated as
    immutable).

    Faster than ``copy.deepcopy`` for JSON Schema documents.

    :raises ValueError: if the object exceeds 255 levels of nesting.
    """
    ...



import json
import socket
import subprocess
import sys
import textwrap
import time
import pytest
from jsonschema_rs import (
    Draft4Validator,
    Draft6Validator,
    Draft7Validator,
    Draft201909Validator,
    Draft202012Validator,
    HttpOptions,
    ValidationError,
    is_valid,
    validator_for,
)


def get_free_port():
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("127.0.0.1", 0))
        return s.getsockname()[1]


@pytest.fixture
def schema_server():
    processes = []

    def start(schema=None):
        if schema is None:
            schema = {"type": "string"}
        port = get_free_port()
        schema_json = json.dumps(schema)

        server_code = textwrap.dedent(f"""
            import json
            import http.server
            import socketserver

            class Handler(http.server.BaseHTTPRequestHandler):
                def do_GET(self):
                    response = {schema_json!r}.encode()
                    self.send_response(200)
                    self.send_header("Content-Type", "application/json")
                    self.send_header("Content-Length", str(len(response)))
                    self.send_header("Connection", "close")
                    self.end_headers()
                    self.wfile.write(response)
                def log_message(self, *args):
                    pass

            with socketserver.TCPServer(("127.0.0.1", {port}), Handler) as httpd:
                print("READY", flush=True)
                httpd.serve_forever()
        """)

        proc = subprocess.Popen(
            [sys.executable, "-c", server_code],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        processes.append(proc)

        # Wait for server to be ready
        for line in iter(proc.stdout.readline, ""):
            if "READY" in line:
                break

        time.sleep(0.05)  # Small delay to ensure server is accepting connections
        return f"http://127.0.0.1:{port}"

    yield start

    for proc in processes:
        proc.terminate()
        proc.wait()


def test_http_options_repr():
    opts = HttpOptions(timeout=30.0, connect_timeout=10.0, tls_verify=False, ca_cert="/cert.pem")
    assert repr(opts) == "HttpOptions(timeout=30, connect_timeout=10, tls_verify=False, ca_cert='/cert.pem')"

    opts2 = HttpOptions(tls_verify=True)
    assert repr(opts2) == "HttpOptions(timeout=None, connect_timeout=None, tls_verify=True, ca_cert=None)"


def test_http_options_equality():
    opts1 = HttpOptions(timeout=30.0, connect_timeout=10.0)
    opts2 = HttpOptions(timeout=30.0, connect_timeout=10.0)
    opts3 = HttpOptions(timeout=30.0, connect_timeout=5.0)

    assert opts1 == opts2
    assert opts1 != opts3
    assert opts1 != "not an HttpOptions"


def test_http_options_hashable():
    opts1 = HttpOptions(timeout=30.0)
    opts2 = HttpOptions(timeout=30.0)
    assert len({opts1, opts2}) == 1


@pytest.mark.parametrize("value", [-1.0, float("nan"), float("inf"), float("-inf")])
def test_http_options_invalid_timeout(value):
    with pytest.raises(ValueError, match="non-negative finite"):
        HttpOptions(timeout=value)


@pytest.mark.parametrize("value", [-1.0, float("nan"), float("inf"), float("-inf")])
def test_http_options_invalid_connect_timeout(value):
    with pytest.raises(ValueError, match="non-negative finite"):
        HttpOptions(connect_timeout=value)


def test_http_options_with_validator_for():
    opts = HttpOptions(timeout=30.0)
    validator = validator_for({"type": "string"}, http_options=opts)
    assert validator.is_valid("test")


def test_http_options_with_is_valid():
    opts = HttpOptions(timeout=30.0)
    assert is_valid({"type": "string"}, "test", http_options=opts)


@pytest.mark.parametrize(
    "validator_cls", [Draft4Validator, Draft6Validator, Draft7Validator, Draft201909Validator, Draft202012Validator]
)
def test_http_options_with_draft_validators(validator_cls):
    opts = HttpOptions(timeout=30.0, tls_verify=False)
    validator = validator_cls({"type": "integer"}, http_options=opts)
    assert validator.is_valid(42)
    assert not validator.is_valid("not an integer")


def test_http_options_invalid_cert_path():
    opts = HttpOptions(ca_cert="/nonexistent/path/to/cert.pem")
    with pytest.raises(RuntimeError, match="Failed to configure HTTP options"):
        validator_for({"$ref": "https://example.com/schema.json"}, http_options=opts)


def test_http_fetch_external_schema():
    # Fetches the JSON Schema 2020-12 meta-schema over HTTPS (requires network)
    schema = {"$ref": "https://json-schema.org/draft/2020-12/schema"}
    validator = validator_for(schema)
    # The meta-schema validates objects
    assert validator.is_valid({})
    assert validator.is_valid({"type": "string"})


def test_http_fetch_schema(schema_server):
    base_url = schema_server()
    schema = {"$ref": f"{base_url}/schema.json"}
    opts = HttpOptions(timeout=5.0, connect_timeout=2.0)
    validator = validator_for(schema, http_options=opts)
    assert validator.is_valid("test")
    assert not validator.is_valid(42)


@pytest.fixture
def slow_server():
    # Server that delays response to test timeout behavior
    processes = []

    def start(delay_seconds):
        port = get_free_port()
        schema_json = json.dumps({"type": "string"})

        server_code = textwrap.dedent(f"""
            import time
            import http.server
            import socketserver

            class Handler(http.server.BaseHTTPRequestHandler):
                def do_GET(self):
                    time.sleep({delay_seconds})
                    response = {schema_json!r}.encode()
                    self.send_response(200)
                    self.send_header("Content-Type", "application/json")
                    self.send_header("Content-Length", str(len(response)))
                    self.send_header("Connection", "close")
                    self.end_headers()
                    self.wfile.write(response)
                def log_message(self, *args):
                    pass

            with socketserver.TCPServer(("127.0.0.1", {port}), Handler) as httpd:
                print("READY", flush=True)
                httpd.serve_forever()
        """)

        proc = subprocess.Popen(
            [sys.executable, "-c", server_code],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        processes.append(proc)

        for line in iter(proc.stdout.readline, ""):
            if "READY" in line:
                break

        time.sleep(0.05)
        return f"http://127.0.0.1:{port}"

    yield start

    for proc in processes:
        proc.terminate()
        proc.wait()


def test_http_timeout_triggers(slow_server):
    # Server delays 2 seconds, but we set timeout to 0.5 seconds
    base_url = slow_server(delay_seconds=2)
    schema = {"$ref": f"{base_url}/schema.json"}
    opts = HttpOptions(timeout=0.5)
    with pytest.raises(ValidationError, match="error sending request"):
        validator_for(schema, http_options=opts)



import pytest

import jsonschema_rs


class EvenValidator:
    def __init__(self, parent_schema, value, schema_path):
        self.enabled = value

    def validate(self, instance):
        if not self.enabled:
            return
        if isinstance(instance, int) and instance % 2 != 0:
            raise ValueError(f"{instance} is not even")


class OddValidator:
    def __init__(self, parent_schema, value, schema_path):
        pass

    def validate(self, instance):
        if isinstance(instance, int) and instance % 2 == 0:
            raise ValueError(f"{instance} is not odd")


class RangeValidator:
    def __init__(self, parent_schema, value, schema_path):
        self.min = value.get("min", float("-inf"))
        self.max = value.get("max", float("inf"))

    def validate(self, instance):
        if isinstance(instance, (int, float)):
            if not (self.min <= instance <= self.max):
                raise ValueError(f"Value {instance} not in range [{self.min}, {self.max}]")


class PositiveValidator:
    def __init__(self, parent_schema, value, schema_path):
        pass

    def validate(self, instance):
        if isinstance(instance, (int, float)) and instance <= 0:
            raise ValueError(f"{instance} is not positive")


@pytest.mark.parametrize(
    "instance, expected",
    [
        (2, True),
        (4, True),
        (0, True),
        (1, False),
        (3, False),
        ("not a number", True),
    ],
)
def test_is_valid(instance, expected):
    validator = jsonschema_rs.validator_for(
        {"even": True},
        keywords={"even": EvenValidator},
    )
    assert validator.is_valid(instance) == expected


@pytest.mark.parametrize(
    "instance, should_raise",
    [
        (2, False),
        (1, True),
        (3, True),
    ],
)
def test_validate(instance, should_raise):
    validator = jsonschema_rs.validator_for(
        {"even": True},
        keywords={"even": EvenValidator},
    )
    if should_raise:
        with pytest.raises(jsonschema_rs.ValidationError):
            validator.validate(instance)
    else:
        validator.validate(instance)


def test_error_message():
    validator = jsonschema_rs.validator_for(
        {"even": True},
        keywords={"even": EvenValidator},
    )
    with pytest.raises(jsonschema_rs.ValidationError) as exc_info:
        validator.validate(3)
    assert "3 is not even" in str(exc_info.value)


def test_iter_errors():
    validator = jsonschema_rs.validator_for(
        {"even": True},
        keywords={"even": EvenValidator},
    )
    errors = list(validator.iter_errors(3))
    assert len(errors) == 1
    assert isinstance(errors[0], jsonschema_rs.ValidationError)


def test_with_standard_keywords():
    validator = jsonschema_rs.validator_for(
        {"type": "integer", "minimum": 0, "even": True},
        keywords={"even": EvenValidator},
    )
    assert validator.is_valid(2)
    assert validator.is_valid(100)
    assert not validator.is_valid(3)
    assert not validator.is_valid(-2)
    assert not validator.is_valid("hello")


def test_multiple_custom_keywords():
    validator = jsonschema_rs.validator_for(
        {"even": True, "isPositive": True},
        keywords={
            "even": EvenValidator,
            "isPositive": PositiveValidator,
        },
    )
    assert validator.is_valid(2)
    assert validator.is_valid(100)
    assert not validator.is_valid(3)
    assert not validator.is_valid(-2)
    assert not validator.is_valid(-3)


def test_nested_schema():
    validator = jsonschema_rs.validator_for(
        {
            "type": "object",
            "properties": {
                "count": {"type": "integer", "even": True},
            },
        },
        keywords={"even": EvenValidator},
    )
    assert validator.is_valid({"count": 2})
    assert validator.is_valid({"count": 100})
    assert not validator.is_valid({"count": 3})


@pytest.mark.parametrize(
    "instance, expected",
    [
        (1, True),
        (3, True),
        (5, True),
        (2, False),
        (4, False),
        ("not a number", True),
    ],
)
def test_odd_validator(instance, expected):
    validator = jsonschema_rs.validator_for(
        {"odd": True},
        keywords={"odd": OddValidator},
    )
    assert validator.is_valid(instance) == expected


@pytest.mark.parametrize(
    "instance, expected",
    [
        (5, True),
        (10, True),
        (0, True),
        (-1, False),
        (11, False),
        (100, False),
    ],
)
def test_range_validator_is_valid(instance, expected):
    validator = jsonschema_rs.validator_for(
        {"customRange": {"min": 0, "max": 10}},
        keywords={"customRange": RangeValidator},
    )
    assert validator.is_valid(instance) == expected


@pytest.mark.parametrize(
    "instance, should_raise",
    [
        (5, False),
        (-1, True),
        (20, True),
    ],
)
def test_range_validator_validate(instance, should_raise):
    validator = jsonschema_rs.validator_for(
        {"customRange": {"min": 0, "max": 10}},
        keywords={"customRange": RangeValidator},
    )
    if should_raise:
        with pytest.raises(jsonschema_rs.ValidationError):
            validator.validate(instance)
    else:
        validator.validate(instance)


def test_standalone_is_valid():
    assert jsonschema_rs.is_valid(
        {"even": True},
        2,
        keywords={"even": EvenValidator},
    )
    assert not jsonschema_rs.is_valid(
        {"even": True},
        3,
        keywords={"even": EvenValidator},
    )


def test_standalone_validate():
    jsonschema_rs.validate(
        {"even": True},
        2,
        keywords={"even": EvenValidator},
    )
    with pytest.raises(jsonschema_rs.ValidationError):
        jsonschema_rs.validate(
            {"even": True},
            3,
            keywords={"even": EvenValidator},
        )


def test_standalone_iter_errors():
    errors = list(
        jsonschema_rs.iter_errors(
            {"even": True},
            3,
            keywords={"even": EvenValidator},
        )
    )
    assert len(errors) == 1
    assert isinstance(errors[0], jsonschema_rs.ValidationError)


def test_standalone_evaluate():
    evaluation = jsonschema_rs.evaluate(
        {"even": True},
        2,
        keywords={"even": EvenValidator},
    )
    result = evaluation.flag()
    assert result["valid"] is True


def test_draft_validators():
    for validator_class in [
        jsonschema_rs.Draft4Validator,
        jsonschema_rs.Draft6Validator,
        jsonschema_rs.Draft7Validator,
        jsonschema_rs.Draft201909Validator,
        jsonschema_rs.Draft202012Validator,
    ]:
        validator = validator_class(
            {"even": True},
            keywords={"even": EvenValidator},
        )
        assert validator.is_valid(2)
        assert not validator.is_valid(3)


def test_receives_correct_arguments():
    received_args = {}

    class CaptureArgs:
        def __init__(self, parent_schema, value, schema_path):
            received_args["parent_schema"] = parent_schema
            received_args["value"] = value
            received_args["schema_path"] = list(schema_path)

        def validate(self, instance):
            pass

    jsonschema_rs.validator_for(
        {"type": "string", "myKeyword": {"foo": "bar"}},
        keywords={"myKeyword": CaptureArgs},
    )

    assert "type" in received_args["parent_schema"]
    assert received_args["parent_schema"]["type"] == "string"
    assert received_args["value"] == {"foo": "bar"}


def test_non_callable_raises_error():
    with pytest.raises(ValueError) as exc_info:
        jsonschema_rs.validator_for(
            {"myKeyword": True},
            keywords={"myKeyword": "not a callable"},
        )
    assert "callable" in str(exc_info.value).lower()


def test_init_error_propagates():
    class FailingInit:
        def __init__(self, parent_schema, value, schema_path):
            raise RuntimeError("Init failed!")

        def validate(self, instance):
            pass

    with pytest.raises(jsonschema_rs.ValidationError) as exc_info:
        jsonschema_rs.validator_for(
            {"failing": True},
            keywords={"failing": FailingInit},
        )

    assert "Init failed" in str(exc_info.value)


class DomainValidationError(Exception):
    """User-defined exception for domain-specific validation failures."""

    def __init__(self, field, reason):
        self.field = field
        self.reason = reason
        super().__init__(f"Field '{field}': {reason}")


def test_custom_exception_class_as_cause():
    class StrictKeyword:
        def __init__(self, parent_schema, value, schema_path):
            pass

        def validate(self, instance):
            if not isinstance(instance, str):
                raise DomainValidationError("value", "must be a string")

    validator = jsonschema_rs.validator_for(
        {"strict": True},
        keywords={"strict": StrictKeyword},
    )
    with pytest.raises(jsonschema_rs.ValidationError) as exc_info:
        validator.validate(123)

    cause = exc_info.value.__cause__
    assert isinstance(cause, DomainValidationError)
    assert cause.field == "value"
    assert cause.reason == "must be a string"
    assert "Field 'value': must be a string" in str(exc_info.value)


def test_validate_cause_is_preserved():
    class FailingValidator:
        def __init__(self, parent_schema, value, schema_path):
            pass

        def validate(self, instance):
            raise ValueError("specific error detail")

    validator = jsonschema_rs.validator_for(
        {"kw": True},
        keywords={"kw": FailingValidator},
    )
    with pytest.raises(jsonschema_rs.ValidationError) as exc_info:
        validator.validate(42)
    assert isinstance(exc_info.value.__cause__, ValueError)
    assert "specific error detail" in str(exc_info.value.__cause__)
    assert exc_info.value.__suppress_context__ is True


def test_iter_errors_cause_is_preserved():
    class FailingValidator:
        def __init__(self, parent_schema, value, schema_path):
            pass

        def validate(self, instance):
            raise TypeError("type-level problem")

    validator = jsonschema_rs.validator_for(
        {"kw": True},
        keywords={"kw": FailingValidator},
    )
    errors = list(validator.iter_errors(42))
    assert len(errors) == 1
    assert isinstance(errors[0].__cause__, TypeError)
    assert "type-level problem" in str(errors[0].__cause__)


def test_keyword_disabled_by_value():
    validator = jsonschema_rs.validator_for(
        {"even": False},
        keywords={"even": EvenValidator},
    )
    assert validator.is_valid(1)
    assert validator.is_valid(2)
    assert validator.is_valid(3)

    validator_enabled = jsonschema_rs.validator_for(
        {"even": True},
        keywords={"even": EvenValidator},
    )
    assert validator_enabled.is_valid(2)
    assert not validator_enabled.is_valid(3)



import json
import os
import pytest
import jsonschema_rs


# Keywords whose annotation is the schema value itself.
_SCHEMA_VALUED_KEYWORDS = frozenset({"contentSchema"})

SUITE_PATH = os.path.join(
    os.path.dirname(__file__),
    "../../jsonschema/tests/suite/annotations/tests",
)

def load_test_cases():
    cases = []
    for filename in sorted(os.listdir(SUITE_PATH)):
        if not filename.endswith(".json"):
            continue
        filepath = os.path.join(SUITE_PATH, filename)
        with open(filepath, encoding="utf-8") as f:
            data = json.load(f)
        for suite_case in data.get("suite", []):
            schema = suite_case["schema"]
            description = suite_case.get("description", filename)
            tests = suite_case.get("tests", [])
            for test in tests:
                instance = test["instance"]
                assertions = test.get("assertions", [])
                cases.append(pytest.param(schema, instance, assertions, id=f"{filename}::{description}"))
    return cases


def get_annotations(schema, instance):
    validator = jsonschema_rs.validator_for(schema)
    evaluation = validator.evaluate(instance)
    raw = evaluation.annotations()

    result = {}
    for ann in raw:
        instance_loc = ann.get("instanceLocation", "")
        annotations = ann.get("annotations", {})

        schema_loc = ann.get("schemaLocation", "")
        segments = [s for s in schema_loc.split("/") if s]
        keyword_from_loc = segments[-1] if segments else None

        if keyword_from_loc in _SCHEMA_VALUED_KEYWORDS:
            # Keep schema-valued annotations as a single value.
            key = (instance_loc, keyword_from_loc)
            result.setdefault(key, []).append(annotations)
        elif isinstance(annotations, dict):
            # Hash bundles are {keyword: value}.
            for keyword, value in annotations.items():
                key = (instance_loc, keyword)
                result.setdefault(key, []).append(value)
        else:
            if keyword_from_loc:
                key = (instance_loc, keyword_from_loc)
                result.setdefault(key, []).append(annotations)
    return result


@pytest.mark.parametrize("schema,instance,assertions", load_test_cases())
def test_annotation(schema, instance, assertions):

    collected = get_annotations(schema, instance)

    for assertion in assertions:
        location = assertion["location"]
        keyword = assertion["keyword"]
        expected = assertion["expected"]

        key = (location, keyword)
        actual_values = collected.get(key, [])

        if not expected:
            assert not actual_values, (
                f"\nKeyword   : {keyword!r}"
                f"\nInstance  : {location!r}"
                f"\nExpected  : no annotation"
                f"\nGot       : {actual_values!r}"
            )
        else:
            for schema_loc, expected_value in expected.items():
                assert expected_value in actual_values, (
                    f"\nKeyword   : {keyword!r}"
                    f"\nInstance  : {location!r}"
                    f"\nSchema    : {schema_loc!r}"
                    f"\nExpected  : {expected_value!r}"
                    f"\nGot       : {actual_values!r}"
                )



import sys

import pytest
from hypothesis import given
from hypothesis import strategies as st

from jsonschema_rs import canonical

clone = canonical.schema.clone


@pytest.mark.parametrize("value", [None, True, False, 0, 1, -7, 3.14, "hello", ""])
def test_immutables_return_same_object(value):
    assert clone(value) is value


def test_empty_dict_is_new_object():
    d = {}
    assert clone(d) is not d
    assert clone(d) == d


def test_empty_list_is_new_object():
    lst = []
    assert clone(lst) is not lst
    assert clone(lst) == lst


def test_dict_is_new_object():
    d = {"a": 1}
    result = clone(d)
    assert result is not d
    assert result == d


def test_list_is_new_object():
    lst = [1, 2, 3]
    result = clone(lst)
    assert result is not lst
    assert result == lst


def test_string_values_are_shared():
    s = "shared"
    d = {"key": s}
    result = clone(d)
    assert result["key"] is s


def test_string_items_in_list_are_shared():
    s = "shared"
    lst = [s, s]
    result = clone(lst)
    assert result[0] is s
    assert result[1] is s


def test_nested_dict_is_deep_cloned():
    inner = {"x": 1}
    outer = {"inner": inner}
    result = clone(outer)
    assert result is not outer
    assert result["inner"] is not inner
    assert result["inner"] == inner


def test_nested_list_is_deep_cloned():
    inner = [1, 2]
    outer = [inner]
    result = clone(outer)
    assert result is not outer
    assert result[0] is not inner
    assert result[0] == inner


def test_complex_schema():
    schema = {
        "type": "object",
        "properties": {
            "name": {"type": "string"},
            "age": {"type": "integer", "minimum": 0},
            "tags": {"type": "array", "items": {"type": "string"}},
        },
        "required": ["name"],
    }
    result = clone(schema)
    assert result == schema
    assert result is not schema
    assert result["properties"] is not schema["properties"]
    assert result["required"] is not schema["required"]
    # string leaves are shared
    assert result["type"] is schema["type"]
    assert result["required"][0] is schema["required"][0]


def test_mutating_clone_does_not_affect_original():
    original = {"a": {"b": 1}}
    cloned = clone(original)
    cloned["a"]["b"] = 999
    assert original["a"]["b"] == 1


def test_mutating_original_does_not_affect_clone():
    original = {"a": [1, 2, 3]}
    cloned = clone(original)
    original["a"].append(4)
    assert cloned["a"] == [1, 2, 3]


def test_tuple_is_returned_as_is():
    t = (1, 2, 3)
    assert clone(t) is t


def test_unknown_object_is_returned_as_is():
    class Custom:
        pass
    obj = Custom()
    assert clone(obj) is obj


def test_recursion_limit_raises():
    value = {}
    node = value
    for _ in range(255):
        inner = {}
        node["x"] = inner
        node = inner
    with pytest.raises(ValueError, match="Recursion limit reached"):
        clone(value)


def test_recursion_limit_does_not_raise_at_254():
    value = {}
    node = value
    for _ in range(254):
        inner = {}
        node["x"] = inner
        node = inner
    # One level short of the limit — must not raise
    clone(value)


_json_scalars = st.one_of(
    st.none(),
    st.booleans(),
    st.integers(),
    st.floats(allow_nan=False, allow_infinity=False),
    st.text(),
)
_json_values = st.recursive(
    _json_scalars,
    lambda children: st.one_of(
        st.lists(children),
        st.dictionaries(st.text(), children),
    ),
    max_leaves=50,
)


@given(_json_values)
def test_clone_equals_original(value):
    assert clone(value) == value


@given(_json_values)
def test_clone_is_independent(value):
    # Cloning a clone should produce equal output
    assert clone(clone(value)) == value


@pytest.mark.skipif(not hasattr(sys, "getrefcount"), reason="CPython only")
def test_refcount_stable_for_string():
    s = "stable"
    d = {"key": s}
    baseline = sys.getrefcount(s)
    for _ in range(100):
        clone(d)
    assert sys.getrefcount(s) == baseline



import copy
import enum
from decimal import Decimal

import pytest

from jsonschema_rs import canonical

to_string = canonical.json.to_string

clone = canonical.schema.clone


def deepclone(value):
    """Specialized deepcopy that copies only dict and list (unrolled 3 levels)."""
    if isinstance(value, dict):
        return {
            k1: (
                {
                    k2: (
                        {k3: deepclone(v3) for k3, v3 in v2.items()}
                        if isinstance(v2, dict)
                        else [deepclone(v3) for v3 in v2]
                        if isinstance(v2, list)
                        else v2
                    )
                    for k2, v2 in v1.items()
                }
                if isinstance(v1, dict)
                else [deepclone(v2) for v2 in v1]
                if isinstance(v1, list)
                else v1
            )
            for k1, v1 in value.items()
        }
    if isinstance(value, list):
        return [
            {k2: deepclone(v2) for k2, v2 in v1.items()}
            if isinstance(v1, dict)
            else [deepclone(v2) for v2 in v1]
            if isinstance(v1, list)
            else v1
            for v1 in value
        ]
    return value


BENCHMARK_CONFIG = {
    "min_rounds": 20,
    "warmup": True,
}


# Reversed insertion order keeps the input intentionally non-canonical
# and exercises key collection + sorting in canonical.json.to_string.
LARGE_DICT_1024 = {f"key-{i:04d}": i for i in range(1023, -1, -1)}


EnumKey = enum.Enum(
    "EnumKey",
    {f"K{i:04d}": f"key-{i:04d}" for i in range(512)},
    type=str,
)
ENUM_KEY_DICT_512 = {member: idx for idx, member in enumerate(list(EnumKey)[::-1])}


INT_FLOAT_LIST_4096 = [float(i) for i in range(4096)]
DECIMAL_FRACTIONAL_LIST_2048 = [Decimal(f"{i}.{(i * 37) % 1000:03d}") for i in range(1, 2049)]
DECIMAL_SPECIAL_LIST_2048 = [Decimal("NaN"), Decimal("Infinity"), Decimal("-Infinity")] * 682 + [Decimal("NaN"), Decimal("Infinity")]


MIXED_NESTED = {
    "meta": {"version": 1, "source": "benchmark", "active": True},
    "items": [
        {
            "id": i,
            "name": f"item-{i}",
            "scores": [float(i), float(i + 1), float(i + 2)],
            "props": {"z": i, "a": i + 1, "m": i + 2},
        }
        for i in range(300)
    ],
}


SMALL_SCHEMA = {
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "age": {"type": "integer", "minimum": 0},
        "active": {"type": "boolean"},
    },
    "required": ["name"],
    "additionalProperties": False,
}

MEDIUM_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "properties": {
        "id": {"type": "integer"},
        "user": {
            "type": "object",
            "properties": {
                "name": {"type": "string", "minLength": 1, "maxLength": 100},
                "email": {"type": "string", "format": "email"},
                "roles": {
                    "type": "array",
                    "items": {"type": "string", "enum": ["admin", "user", "guest"]},
                },
                "address": {
                    "type": "object",
                    "properties": {
                        "street": {"type": "string"},
                        "city": {"type": "string"},
                        "country": {"type": "string", "minLength": 2, "maxLength": 2},
                    },
                },
            },
            "required": ["name", "email"],
        },
        "tags": {"type": "array", "items": {"type": "string"}, "uniqueItems": True},
        "metadata": {"type": "object", "additionalProperties": {"type": "string"}},
    },
    "required": ["id", "user"],
}

_prop = lambda t: {"type": t}
LARGE_SCHEMA = {
    "type": "object",
    "properties": {
        f"field_{i}": {
            "type": "object",
            "properties": {
                "value": _prop("string"),
                "count": _prop("integer"),
                "enabled": _prop("boolean"),
                "tags": {"type": "array", "items": _prop("string")},
                "nested": {
                    "type": "object",
                    "properties": {
                        "a": _prop("string"),
                        "b": _prop("integer"),
                        "c": {"type": "array", "items": _prop("number")},
                    },
                },
            },
        }
        for i in range(50)
    },
}


@pytest.mark.benchmark(group="canonical-json-dict-sort", **BENCHMARK_CONFIG)
def test_canonical_json_large_dict_1024(benchmark):
    benchmark(to_string, LARGE_DICT_1024)


@pytest.mark.benchmark(group="canonical-json-enum-keys", **BENCHMARK_CONFIG)
def test_canonical_json_enum_key_dict_512(benchmark):
    benchmark(to_string, ENUM_KEY_DICT_512)


@pytest.mark.benchmark(group="canonical-json-int-floats", **BENCHMARK_CONFIG)
def test_canonical_json_int_float_list_4096(benchmark):
    benchmark(to_string, INT_FLOAT_LIST_4096)


@pytest.mark.benchmark(group="canonical-json-decimal-fractional", **BENCHMARK_CONFIG)
def test_canonical_json_decimal_fractional_list_2048(benchmark):
    benchmark(to_string, DECIMAL_FRACTIONAL_LIST_2048)


@pytest.mark.benchmark(group="canonical-json-decimal-special", **BENCHMARK_CONFIG)
def test_canonical_json_decimal_special_list_2048(benchmark):
    benchmark(to_string, DECIMAL_SPECIAL_LIST_2048)


@pytest.mark.benchmark(group="canonical-json-mixed-nested", **BENCHMARK_CONFIG)
def test_canonical_json_mixed_nested(benchmark):
    benchmark(to_string, MIXED_NESTED)


_CLONE_IMPLS = [
    pytest.param(copy.deepcopy, id="deepcopy"),
    pytest.param(deepclone, id="deepclone"),
    pytest.param(clone, id="canonical"),
]


@pytest.mark.parametrize("fn", _CLONE_IMPLS)
@pytest.mark.benchmark(group="clone-small", **BENCHMARK_CONFIG)
def test_clone_small(benchmark, fn):
    benchmark(fn, SMALL_SCHEMA)


@pytest.mark.parametrize("fn", _CLONE_IMPLS)
@pytest.mark.benchmark(group="clone-medium", **BENCHMARK_CONFIG)
def test_clone_medium(benchmark, fn):
    benchmark(fn, MEDIUM_SCHEMA)


@pytest.mark.parametrize("fn", _CLONE_IMPLS)
@pytest.mark.benchmark(group="clone-large", **BENCHMARK_CONFIG)
def test_clone_large(benchmark, fn):
    benchmark(fn, LARGE_SCHEMA)



import jsonschema_rs


def test_evaluate_produces_expected_outputs_for_valid_instance():
    schema = {
        "type": "object",
        "properties": {"name": {"type": "string"}, "age": {"type": "number", "minimum": 0}},
        "required": ["name"],
    }
    instance = {"name": "Alice", "age": 1}

    evaluation = jsonschema_rs.evaluate(schema, instance)

    assert evaluation.flag() == {"valid": True}

    assert evaluation.list() == {
        "valid": True,
        "details": [
            {
                "evaluationPath": "",
                "instanceLocation": "",
                "schemaLocation": "",
                "valid": True,
            },
            {
                "valid": True,
                "evaluationPath": "/type",
                "instanceLocation": "",
                "schemaLocation": "/type",
            },
            {
                "valid": True,
                "evaluationPath": "/required",
                "instanceLocation": "",
                "schemaLocation": "/required",
            },
            {
                "valid": True,
                "evaluationPath": "/properties",
                "instanceLocation": "",
                "schemaLocation": "/properties",
                "annotations": ["age", "name"],
            },
            {
                "valid": True,
                "evaluationPath": "/properties/age",
                "instanceLocation": "/age",
                "schemaLocation": "/properties/age",
            },
            {
                "valid": True,
                "evaluationPath": "/properties/age/type",
                "instanceLocation": "/age",
                "schemaLocation": "/properties/age/type",
            },
            {
                "valid": True,
                "evaluationPath": "/properties/age/minimum",
                "instanceLocation": "/age",
                "schemaLocation": "/properties/age/minimum",
            },
            {
                "valid": True,
                "evaluationPath": "/properties/name",
                "instanceLocation": "/name",
                "schemaLocation": "/properties/name",
            },
            {
                "valid": True,
                "evaluationPath": "/properties/name/type",
                "instanceLocation": "/name",
                "schemaLocation": "/properties/name/type",
            },
        ],
    }

    assert evaluation.hierarchical() == {
        "valid": True,
        "evaluationPath": "",
        "instanceLocation": "",
        "schemaLocation": "",
        "details": [
            {
                "valid": True,
                "evaluationPath": "/type",
                "instanceLocation": "",
                "schemaLocation": "/type",
            },
            {
                "valid": True,
                "evaluationPath": "/required",
                "instanceLocation": "",
                "schemaLocation": "/required",
            },
            {
                "valid": True,
                "evaluationPath": "/properties",
                "instanceLocation": "",
                "schemaLocation": "/properties",
                "annotations": ["age", "name"],
                "details": [
                    {
                        "valid": True,
                        "evaluationPath": "/properties/age",
                        "instanceLocation": "/age",
                        "schemaLocation": "/properties/age",
                        "details": [
                            {
                                "valid": True,
                                "evaluationPath": "/properties/age/type",
                                "instanceLocation": "/age",
                                "schemaLocation": "/properties/age/type",
                            },
                            {
                                "valid": True,
                                "evaluationPath": "/properties/age/minimum",
                                "instanceLocation": "/age",
                                "schemaLocation": "/properties/age/minimum",
                            },
                        ],
                    },
                    {
                        "valid": True,
                        "evaluationPath": "/properties/name",
                        "instanceLocation": "/name",
                        "schemaLocation": "/properties/name",
                        "details": [
                            {
                                "valid": True,
                                "evaluationPath": "/properties/name/type",
                                "instanceLocation": "/name",
                                "schemaLocation": "/properties/name/type",
                            }
                        ],
                    },
                ],
            },
        ],
    }

    assert evaluation.annotations() == [
        {
            "schemaLocation": "/properties",
            "absoluteKeywordLocation": None,
            "instanceLocation": "",
            "annotations": ["age", "name"],
        }
    ]
    assert evaluation.errors() == []


def test_validator_evaluate_annotations_and_errors():
    schema = {
        "type": "array",
        "prefixItems": [{"type": "string"}],
        "items": {"type": "integer"},
    }
    validator = jsonschema_rs.validator_for(schema)

    valid_eval = validator.evaluate(["hello", 1])
    assert valid_eval.annotations() == [
        {
            "schemaLocation": "/items",
            "absoluteKeywordLocation": None,
            "instanceLocation": "",
            "annotations": True,
        },
        {
            "schemaLocation": "/prefixItems",
            "absoluteKeywordLocation": None,
            "instanceLocation": "",
            "annotations": 0,
        },
    ]
    assert valid_eval.errors() == []

    invalid_eval = validator.evaluate(["hello", "oops"])
    assert invalid_eval.flag() == {"valid": False}
    assert invalid_eval.errors() == [
        {
            "schemaLocation": "/items/type",
            "absoluteKeywordLocation": None,
            "instanceLocation": "/1",
            "error": '"oops" is not of type "integer"',
        }
    ]
    assert invalid_eval.list() == {
        "valid": False,
        "details": [
            {
                "evaluationPath": "",
                "instanceLocation": "",
                "schemaLocation": "",
                "valid": False,
            },
            {
                "valid": True,
                "evaluationPath": "/type",
                "instanceLocation": "",
                "schemaLocation": "/type",
            },
            {
                "valid": False,
                "evaluationPath": "/items",
                "instanceLocation": "",
                "schemaLocation": "/items",
                "droppedAnnotations": True,
            },
            {
                "valid": False,
                "evaluationPath": "/items",
                "instanceLocation": "/1",
                "schemaLocation": "/items",
            },
            {
                "valid": False,
                "evaluationPath": "/items/type",
                "instanceLocation": "/1",
                "schemaLocation": "/items/type",
                "errors": {"type": '"oops" is not of type "integer"'},
            },
            {
                "valid": True,
                "evaluationPath": "/prefixItems",
                "instanceLocation": "",
                "schemaLocation": "/prefixItems",
                "annotations": 0,
            },
            {
                "valid": True,
                "evaluationPath": "/prefixItems/0",
                "instanceLocation": "/0",
                "schemaLocation": "/prefixItems/0",
            },
            {
                "valid": True,
                "evaluationPath": "/prefixItems/0/type",
                "instanceLocation": "/0",
                "schemaLocation": "/prefixItems/0/type",
            },
        ],
    }
    assert invalid_eval.hierarchical() == {
        "valid": False,
        "evaluationPath": "",
        "instanceLocation": "",
        "schemaLocation": "",
        "details": [
            {
                "valid": True,
                "evaluationPath": "/type",
                "instanceLocation": "",
                "schemaLocation": "/type",
            },
            {
                "valid": False,
                "evaluationPath": "/items",
                "instanceLocation": "",
                "schemaLocation": "/items",
                "droppedAnnotations": True,
                "details": [
                    {
                        "valid": False,
                        "evaluationPath": "/items",
                        "instanceLocation": "/1",
                        "schemaLocation": "/items",
                        "details": [
                            {
                                "valid": False,
                                "evaluationPath": "/items/type",
                                "instanceLocation": "/1",
                                "schemaLocation": "/items/type",
                                "errors": {"type": '"oops" is not of type "integer"'},
                            }
                        ],
                    }
                ],
            },
            {
                "valid": True,
                "evaluationPath": "/prefixItems",
                "instanceLocation": "",
                "schemaLocation": "/prefixItems",
                "annotations": 0,
                "details": [
                    {
                        "valid": True,
                        "evaluationPath": "/prefixItems/0",
                        "instanceLocation": "/0",
                        "schemaLocation": "/prefixItems/0",
                        "details": [
                            {
                                "valid": True,
                                "evaluationPath": "/prefixItems/0/type",
                                "instanceLocation": "/0",
                                "schemaLocation": "/prefixItems/0/type",
                            }
                        ],
                    }
                ],
            },
        ],
    }
